package debug;

/*
  Debug.java
   - milan ramaiya - 02.03.02

   This class is used for easily enabling/disabling debugging output.

   External calls that can be made are:

     off()  disable output (feeds to NULL stream)
     on()   enable output to standard error
     on(String filename)
     on(String filename, boolean append) 
     on(PrintStream ps)  set custom output stream,
     on(OutputStream ps)  set custom output stream,                  

     print(any)   prints the data as in Java. Also, if an array, formats the
                  elements in human-readable format
     println(any) essentially calls print(any); println(any)

   ------------------------------------------------------------------HISTORY--
   1.0.2   3.17.02  removed enable()/disable/setOutputStream()
                    added off()/on()
                    made class non-static
   1.0.1   3.14.02  added enable()/disable()/setOutputStream()
   1.0.0   2.03.02  original version
*/

import java.io.*;

public class Debug 
{   
  // basic streaming classes
  private static class NullOutputStream extends OutputStream {
      public void write(int b) {}
  }
  
  private static class NullStream extends PrintStream {
    NullStream() {
      super(new NullOutputStream());
    }
  }

  private static class FileStream extends PrintStream {
    FileStream(String filename) throws FileNotFoundException{
      super(new FileOutputStream(filename));
    }
  }
  
  // static constants for streams
  public static PrintStream NULL   = new NullStream();
  public static PrintStream STDERR = System.err;
  public static PrintStream STDOUT = System.out;
  
  private static PrintStream err = STDERR;

  /*
  public Debug()                   { on(); }
  public Debug(OutputStream os)    { on(new PrintStream(os)); }
  public Debug(PrintStream ps)     { err = ps; }
  public Debug(Object o)           { if(o==null) off(); else on(); }
  public Debug(String filename) throws FileNotFoundException
  { on(filename); }  
  */
  
  public static void off()                { on(NULL);   }
  public static void on()                 { on(STDERR); }
  public static void on(OutputStream os)  { on(new PrintStream(os)); }
  public static void on(PrintStream ps)   { err = ps; }
  public static void on(String filename) throws FileNotFoundException
  { on(new FileStream(filename)); }
  
  public class ASSERT_EXCEPTION extends RuntimeException {}

  // output functions
  public static void print(int i)       { err.print(i);    }
  public static void print(char c)      { err.print(c);    }
  public static void print(long l)      { err.print(l);    }
  public static void print(float f)     { err.print(f);    }
  public static void print(double d)    { err.print(d);    }
  public static void print(Object o)    { err.print(o);    }
  public static void print(boolean b)   { err.print(b);    }
  public static void println()          { err.println();   }
  public static void println(int i)     { err.println(i);  }
  public static void println(char c)    { err.println(c);  }
  public static void println(long l)    { err.println(l);  }
  public static void println(float f)   { err.println(f);  }
  public static void println(double d)  { err.println(d);  }
  public static void println(Object o)  { err.println(o);  }
  public static void println(boolean b) { err.println(b);  }

  // array output
  public static void println(int     a[]) { print(a); println(); }
  public static void println(long    a[]) { print(a); println(); }
  public static void println(float   a[]) { print(a); println(); }
  public static void println(double  a[]) { print(a); println(); }
  public static void println(Object  a[]) { print(a); println(); }
  public static void println(boolean a[]) { print(a); println(); }

  public static void print(int a[])
  {
    print("{");
    for(int i=0; i<a.length; i++)
      print(a[i] + (i!=a.length-1 ? ", " : ""));
    print("}");
  }

  public static void print(char a[])
  {
    print("{");
    for(int i=0; i<a.length; i++)
      print(a[i] + (i!=a.length-1 ? ", " : ""));
    print("}");
  }

  public static void print(long a[])
  {
    print("{");
    for(int i=0; i<a.length; i++)
      print(a[i] + (i!=a.length-1 ? ", " : ""));
    print("}");
  }

  public static void print(float a[])
  {
    print("{");
    for(int i=0; i<a.length; i++)
      print(a[i] + (i!=a.length-1 ? ", " : ""));
    print("}");
  }

  public static void print(double a[])
  {
    print("{");
    for(int i=0; i<a.length; i++)
      print(a[i] + (i!=a.length-1 ? ", " : ""));
    print("}");
  }

  public static void print(Object a[])
  {
    print("{");
    for(int i=0; i<a.length; i++)
      print(a[i] + (i!=a.length-1 ? ", " : ""));
    print("}");
  }

  public static void print(boolean a[])
  {
    print("{");
    for(int i=0; i<a.length; i++)
      print(a[i] + (i!=a.length-1 ? ", " : ""));
    print("}");
  }    
}
