package parser;

import java_cup.runtime.Symbol;
import java.io.InputStream;
import error.ErrorMsg;

%%
%class TigerScanner
%implements java_cup.runtime.Scanner
%function next_token
%type java_cup.runtime.Symbol
%char

%{

private ErrorMsg errMsg;
private String string;   // strings between quotes
private int ncomment;    // comment level for nested comments

public TigerScanner(InputStream is, ErrorMsg e) {
  this(is);
  errMsg=e;
}

private void newline() 
{ errMsg.newline(yychar); }

private java_cup.runtime.Symbol tok(int k) 
{ return tok(k, null); }

private java_cup.runtime.Symbol tok(int k, Object value) 
{ return new java_cup.runtime.Symbol(k, yychar, yylength(), value); }  

private void err(String errdesc) 
{ errMsg.error(yychar, errdesc); }

// General functions for string escaping
String repeat(char c, int n)
{
    String s = "";
    for(int i=0; i<n; i++)
        s+=c;
    return s;
}
    
String nFormat(int n, int spc) 
{
    String s=Integer.toString(n);
    return repeat('0', spc-s.length()) + s;
}

// fixes escape codes, assumes correct format
// if printable, shows actual char, otherwise shows escaped char
String escStr(String s)
{
    int n=Integer.parseInt(s.substring(1));
    if(n<32 || n>127)
        return "\\" + nFormat(n,3);
    else
        return (char)n+"";
}

// fixes control codes, assumes correct format
String escCtrl(String s)
{ return "\\" + nFormat(s.charAt(2)-64,3); }

%}

%eofval{
  return tok(TokenConstants.EOF, null);
%eofval}

d   = ([0-9])                        /* digit                 */
ds  = ({d}+)                         /* digits                */
n   = ({ds}|{d}*"."{ds}|{ds}"."{d}*) /* int or float          */
c   = ([a-zA-Z]|_)                   /* alpha character (+ _) */
cd  = ({c}|{d})                      /* character or digit    */
id  = ({c}+{cd}*)                    /* id                    */
ws  = ([" "\r\t\f])                  /* whitespace except \n  */

%state COMMENT STRING STRING_ML

%%

<YYINITIAL>{ws}     {}
<YYINITIAL>\n       { newline(); }

<YYINITIAL>function { return tok(TokenConstants.FUNCTION); }
<YYINITIAL>break    { return tok(TokenConstants.BREAK); }
<YYINITIAL>array    { return tok(TokenConstants.ARRAY); }
<YYINITIAL>while    { return tok(TokenConstants.WHILE); }
<YYINITIAL>else     { return tok(TokenConstants.ELSE); }
<YYINITIAL>then     { return tok(TokenConstants.THEN); }
<YYINITIAL>type     { return tok(TokenConstants.TYPE); }
<YYINITIAL>end      { return tok(TokenConstants.END); }
<YYINITIAL>for      { return tok(TokenConstants.FOR); }
<YYINITIAL>let      { return tok(TokenConstants.LET); }
<YYINITIAL>nil      { return tok(TokenConstants.NIL); }
<YYINITIAL>var      { return tok(TokenConstants.VAR); }
<YYINITIAL>do       { return tok(TokenConstants.DO); }
<YYINITIAL>if       { return tok(TokenConstants.IF); }
<YYINITIAL>in       { return tok(TokenConstants.IN); }
<YYINITIAL>of       { return tok(TokenConstants.OF); }
<YYINITIAL>to       { return tok(TokenConstants.TO); }
<YYINITIAL>{n}      { return tok(TokenConstants.INT, yytext()); }
<YYINITIAL>{id}     { return tok(TokenConstants.ID, yytext()); }
<YYINITIAL>"/*"     { ncomment = 1; yybegin(COMMENT);}
<YYINITIAL>"*/"     { err("Comment end without beginning"); }
<YYINITIAL>":="     { return tok(TokenConstants.ASSIGN); }
<YYINITIAL>"<="     { return tok(TokenConstants.LE); }
<YYINITIAL>">="     { return tok(TokenConstants.GE); }
<YYINITIAL>"<>"     { return tok(TokenConstants.NEQ); }

<YYINITIAL>"<"      { return tok(TokenConstants.LT); }
<YYINITIAL>">"      { return tok(TokenConstants.GT); }
<YYINITIAL>"|"      { return tok(TokenConstants.OR); }
<YYINITIAL>"&"      { return tok(TokenConstants.AND); }
<YYINITIAL>"*"      { return tok(TokenConstants.TIMES); }
<YYINITIAL>"/"      { return tok(TokenConstants.DIVIDE); }
<YYINITIAL>"+"      { return tok(TokenConstants.PLUS); }
<YYINITIAL>"-"      { return tok(TokenConstants.MINUS); }
<YYINITIAL>":"      { return tok(TokenConstants.COLON); }
<YYINITIAL>";"      { return tok(TokenConstants.SEMICOLON); }
<YYINITIAL>","      { return tok(TokenConstants.COMMA); }
<YYINITIAL>"="      { return tok(TokenConstants.EQ); }
<YYINITIAL>"{"      { return tok(TokenConstants.LBRACE); }
<YYINITIAL>"}"      { return tok(TokenConstants.RBRACE); }
<YYINITIAL>"["      { return tok(TokenConstants.LBRACK); }
<YYINITIAL>"]"      { return tok(TokenConstants.RBRACK); }
<YYINITIAL>"("      { return tok(TokenConstants.LPAREN); }
<YYINITIAL>")"      { return tok(TokenConstants.RPAREN); }
<YYINITIAL>"."      { return tok(TokenConstants.DOT); }
<YYINITIAL>\"       { string=""; yybegin(STRING); }

<STRING>\\\n        { newline(); }
<STRING>\\{ws}      { yybegin(STRING_ML); }
<STRING>\\\"        { string+="\\042"; }
<STRING>\\n         { string+="\\012"; }
<STRING>\\t         { string+="\\011"; }
<STRING>\\\\        { string+="\\\\";  }
<STRING>\\(([01][0-9][0-9])|2([0-4][0-9]|5[0-5]))
                    { string+=escStr(yytext()); }
<STRING>\\"^"[@-_]  { string+=escCtrl(yytext()); }
<STRING>\\{d}{d}{d} { err("Decimal escape code '"+yytext().substring(1) + 
                      "' out of range");}
<STRING>\\"^".      { err("Invalid control code '"+yytext().charAt(2)+"'.");}
<STRING>\\.         { err("Invalid escape code: '"+yytext().charAt(1)+"'."); }
<STRING>\"          { yybegin(YYINITIAL);
                      return tok(TokenConstants.STRING, string); }
<STRING>.           { string+=yytext(); }

<STRING_ML>\\       { yybegin(STRING); }
<STRING_ML>{ws}     {}
<STRING_ML>.        { err("Illegal character: expecting '\\'."); }

<COMMENT>"/*"       { ncomment++; }
<COMMENT>"*/"       { ncomment--; if(ncomment==0) yybegin(YYINITIAL);}
<COMMENT>\n         { newline();}
<COMMENT>.|{ws}     {}

\n                  { newline();}
.                   { err("Illegal character: '"+yytext()+"'."); }
