package syntax;

import java.io.PrintStream;
import java.util.Iterator;

public class Print {

   final PrintStream out;

   public Print (PrintStream o) {out=o;}

   void indent (int in) {
      for (int i=0; i<in; i++) out.print(' ');
   }

   void say (String s) {
      out.print(s);
   }

   void say (int i) {
     Integer local = new Integer(i);
     out.print(local.toString());
   }

   void say(boolean b) {
      Boolean local = new Boolean(b);
      out.print(local.toString());
   }
   
   void sayln () {
      say("\n");
   }

   void sayln (String s) {
      say (s); say ("\n");
   }

   void prVar (SimpleVar v, int in) {
      say ("SimpleVar(\""); say (v.name); say ("\")");
   }
  
   void prVar(FieldVar v, int in) {
      sayln ("FieldVar(");
      prVar (v.var, in+2); sayln (",");
      indent (in+2); say (v.field);
      say(")");
   }

   void prVar (SubscriptVar v, int in) {
      sayln ("SubscriptVar(");
      prVar (v.var, in+2); sayln(",");
      prExp (v.index, in+2);
      say (")");
   }

   void prVar (Var v, int in) {
      indent(in);

      /*  Should not be null, but make this more robust. */
      if (v==null) {
	 say ("null Var");
         return;
      }

      if (v instanceof SimpleVar)         prVar((SimpleVar) v, in);
      else if (v instanceof FieldVar)     prVar((FieldVar) v, in);
      else if (v instanceof SubscriptVar) prVar((SubscriptVar) v, in);
      else throw new RuntimeException (v.getClass().getName());
   }
  
   void prExp(OpExp e, int in) {
      sayln ("OpExp(");
      indent (in+2);

      switch (e.oper) {
      case OpExp.PLUS: say("PLUS"); break;
      case OpExp.MINUS: say("MINUS"); break;
      case OpExp.TIMES: say("TIMES"); break;
      case OpExp.DIV:  say("DIV"); break;
      case OpExp.EQ:   say("EQ"); break;
      case OpExp.NE:   say("NE"); break;
      case OpExp.LT: say("LT"); break;
      case OpExp.LE: say("LE"); break;
      case OpExp.GT: say("GT"); break;
      case OpExp.GE: say("GE"); break;
      default:	 throw new RuntimeException ("unknown op="+e.oper);
      }

      sayln (",");
      prExp (e.left, in+2);  sayln (",");
      prExp (e.right, in+2); say (")");
   }

   void prExp (VarExp e, int in) {
      sayln ("VarExp("); prVar (e.var, in+2); say(")");
   }

   void prExp (NilExp e, int in) {
      say ("NilExp()");
   }

   void prExp(IntExp e, int in) {
      say ("IntExp("); say (e.value); say (")");
   }

   void prExp (StringExp e, int d) {
      say("StringExp (\""); say(e.value); say("\")");
   }

   void prExp (CallExp e, int in) {
      say ("CallExp(function name=\""+e.func+"\"");
      for (Iterator i=e.args.iterator(); i.hasNext(); ) {
	 if (i.hasNext()) sayln (",");
	 prExp ((Exp) i.next(), in+2);
      }
      say (")");
   }

   void prExp (RecordExp e, int in) {
      say ("RecordExp(type=\""+e.type+"\"");
      for (Iterator i=e.fields.iterator(); i.hasNext(); ) {
	 if (i.hasNext()) sayln(",");
	 prDec ((Dec) i.next(), in+2);
      }
      say (")");
   }

   void prExp (SeqExp e, int d) {
      if (e.list.size()==0) {
	 say("SeqExp()");
      } else {
	 sayln("SeqExp(");
	 for (Iterator i=e.list.iterator(); i.hasNext(); ) {
	    prExp ((Exp) i.next(), d+2);
	    if (i.hasNext()) sayln(";");
	 }
	 say (")");
      }
   }

   void prExp (AssignExp e, int in) {
      sayln ("AssignExp(");
      prVar (e.var, in+2); sayln(",");
      prExp (e.exp, in+2); say(")");
   }
  
   void prExp (IfExp e, int in) {
      sayln ("IfExp(");
      prExp (e.test, in+2); sayln(",");
      indent (in+2);  sayln ("// then clause");
      prExp (e.thenclause, in+2);
      if (e.elseclause!=null) { /* else is optional */
	 sayln (",");
	 indent (in+2);  sayln ("// else clause");
	 prExp (e.elseclause, in+2);
      }
      say(")");
  }

   void prExp (WhileExp e, int in) {
      sayln ("WhileExp(");
      prExp (e.test, in+2); sayln(",");
      prExp (e.body, in+2);
      say (")");
   }

   void prExp (ForExp e, int in) {
      sayln ("ForExp("); 
      indent (in+2);
      prDec (e.var, in+2); sayln (",");
      prExp (e.hi,  in+2); sayln (",");
      prExp (e.body, in+2);
      say(")");
   }

   void prExp (BreakExp e, int d) {
      say("BreakExp()");
   }


   void prExp (LetExp e, int in) {
      sayln ("LetExp(");
      if (e.decs.size()==0) {
	 indent (in+2); sayln ("List(),");
      } else {
	 indent (in+2); sayln ("List(");
	 for (Iterator i=e.decs.iterator(); i.hasNext(); ) {
	    prDec ((Dec) i.next(), in+4);
	    if (i.hasNext()) sayln(",");
	 }
	 sayln ();
	 indent (in+2); sayln ("),");
      }
      if (e.body.size()==0) {
	 indent (in+2); say ("List()");
      } else {
	 indent (in+2); sayln ("List(");
	 for (Iterator i=e.body.iterator(); i.hasNext(); ) {
	    prExp ((Exp) i.next(), in+4);
	    if (i.hasNext()) sayln(",");
	 }
	 sayln ();
	 indent (in+2); say (")");
      }
      say (")");
   }

   void prExp (ArrayExp e, int in) {
      say ("ArrayExp("); say(e.type); sayln(",");
      prExp(e.size, in); sayln(",");
      prExp(e.init, in); say(")");
  }

   public void prExp (Exp e) {
      prExp (e, 0); sayln ();
   }

   /* Print Exp class types. Indent d spaces; leave at end of line. */
   public void prExp (Exp e, int d) {

      indent(d);

      /*  Should not be null, but make this more robust. */
      if (e==null) {
	 say ("null Exp");
         return;
      }

      if (e instanceof OpExp) prExp((OpExp)e, d);
      else if (e instanceof VarExp) prExp((VarExp) e, d);
      else if (e instanceof NilExp) prExp((NilExp) e, d);
      else if (e instanceof IntExp) prExp((IntExp) e, d);
      else if (e instanceof StringExp) prExp((StringExp) e, d);
      else if (e instanceof CallExp) prExp((CallExp) e, d);
      else if (e instanceof RecordExp) prExp((RecordExp) e, d);
      else if (e instanceof SeqExp) prExp((SeqExp) e, d);
      else if (e instanceof AssignExp) prExp((AssignExp) e, d);
      else if (e instanceof IfExp) prExp((IfExp) e, d);
      else if (e instanceof WhileExp) prExp((WhileExp) e, d);
      else if (e instanceof ForExp) prExp((ForExp) e, d);
      else if (e instanceof BreakExp) prExp((BreakExp) e, d);
      else if (e instanceof LetExp) prExp((LetExp) e, d);
      else if (e instanceof ArrayExp) prExp((ArrayExp) e, d);
      else throw new RuntimeException (e.getClass().getName());
  }

   void prDec (FunctionDec d, int in) {
      sayln("FunctionDec(");
      indent (in+2); sayln ("function name=\""+d.name+"\",");

      // Formal parameters (if any)
      if (d.params.size() != 0) {
	 indent (in+2); sayln ("// formal parameters");
	 for (Iterator i=d.params.iterator(); i.hasNext(); ) {
	    prDec ((Dec) i.next(), in+2);
	    if (i.hasNext()) sayln(",");
	 }
	 sayln(",");
      } else {
	 indent (in+2); sayln ("// no formal parameters");
      }

      if (d.result!=null) {
	 indent (in+2); sayln ("result type=\""+d.result+"\",");
      }

      indent (in+2); sayln ("// function body");

      prExp (d.body, in+2);

      if (d.next!=null) {
	 sayln (",");
	 indent (in+2);  sayln ("// next mutually recursive function declaration");
	 indent (in+2);
	 prDec (d.next, in+2);
      }
      say(")");
   }

   void prDec (VarDec d, int in) {
      sayln ("VarDec(");
      indent (in+2); sayln ("var name=\""+d.name+"\",");
      if (d.type!=null) {
	 if (d.init!=null) {
	    indent(in+2); sayln ("type=\""+d.type+"\",");
	 } else {
	    indent(in+2); say ("type=\""+d.type+"\"");
	 }
      }
      if (d.init!=null) {
	 prExp (d.init, in+2);
      }
      say(")");
   }

   void prDec (TypeDec d, int in) {
      sayln ("TypeDec(");
      indent (in+2); sayln ("type name=\""+d.name+"\",");
      prType (d.type, in+2); 
      if (d.next!=null) {
	 sayln (",");
	 indent (in+2);  sayln ("// next mutually recursive type declaration");
	 indent (in+2);
	 prDec (d.next, in+2); 
      }
      say(")");
   }

   void prDec (Dec d, int in) {
      indent(in);

      /*  Should not be null, but make this more robust. */
      if (d==null) {
	 say ("null Dec");
         return;
      }

      if (d instanceof FunctionDec)  prDec((FunctionDec) d, in);
      else if (d instanceof VarDec)  prDec((VarDec) d, in);
      else if (d instanceof TypeDec) prDec((TypeDec) d, in);
      else throw new RuntimeException (d.getClass().getName());
  }

   void prType (NameType t, int in) {
      say("NameType("); say(t.name); say(")");
   }


   void prType (RecordType t, int in) {
      // Fields
      if (t.fields.size()==0) {
	 say ("RecordType()");
      } else {
	 sayln("RecordType(");
	 for (Iterator i=t.fields.iterator(); i.hasNext(); ) {
	    prDec ((Dec) i.next(), in+2);
	    if (i.hasNext()) sayln(",");
	 }
	 say(")");
      }
   }

   void prType (ArrayType t, int in) {
      say ("ArrayType(\""); say(t.type); say("\")");
   }

   /*  Print types on one line  */
   void prType (TypeX t, int in) {

      indent(in);

      /*  Should not be null, but make this more robust. */
      if (t==null) {
	 say ("null Type");
         return;
      }

      if (t instanceof NameType)        prType ((NameType) t, in);
      else if (t instanceof RecordType) prType ((RecordType) t, in);
      else if (t instanceof ArrayType)  prType((ArrayType) t, in);
      else throw new RuntimeException (t.getClass().getName());
   }
}



