package translate;

import java.util.LinkedList;

import tree.BINOP;
import tree.CALL;
import tree.CJUMP;
import tree.CONST;
import tree.ESEQ;
//import tree.Exp;
import tree.EXP;
import tree.ExpList;
import tree.JUMP;
import tree.LABEL;
import tree.MEM;
import tree.MOVE;
import tree.NAME;
import tree.NOP;
import tree.Printer;
import tree.Print;
import tree.SEQ;
//import tree.Stm;
import tree.StmList;
import tree.TEMP;

import temp.Label;
import temp.Temp;

import frame.Frame;
import compiler.Compiler;

public class Translate 
{

    private static final Frame frame = Compiler.frame;
    private static final int wordsize = frame.wordSize();
    
    public final static int PLUS    = BINOP.PLUS;
    public final static int MINUS   = BINOP.MINUS;
    public final static int TIMES   = BINOP.MUL;
    public final static int MULT    = BINOP.MUL;
    public final static int DIV     = BINOP.DIV;
    public final static int AND     = BINOP.AND;
    public final static int OR      = BINOP.OR;
    public final static int LSHIFT  = BINOP.LSHIFT;
    public final static int RSHIFT  = BINOP.RSHIFT;
    public final static int ARSHIFT = BINOP.ARSHIFT;
    public final static int XOR     = BINOP.XOR;

    public final static int LT  = CJUMP.LT;
    public final static int GT  = CJUMP.GT;
    public final static int LE  = CJUMP.LE;
    public final static int GE  = CJUMP.GE;
    public final static int NE  = CJUMP.NE;
    public final static int EQ  = CJUMP.EQ;
    public final static int ULT = CJUMP.ULT;
    public final static int UGT = CJUMP.UGT;
    public final static int ULE = CJUMP.ULE;
    public final static int UGE = CJUMP.UGE;  
  
    private Fragment frags = new Fragment();

    // ** NOTE TO SELF:
    // NO PRIVATE UTILITY FUNCTIONS SHOULD UTILIZE translate.Exp ** EVER
    // BECAUSE IT MAKES IT ALL THE MORE CONFUSING, (not that it's not already)
    private ESEQ  eseq(tree.Stm s, tree.Exp f) { return new ESEQ(s, f); }
    private ESEQ  eseq(tree.Exp s, tree.Exp f) { return eseq(new EXP(s), f); }
    private MEM   mem(tree.Exp e)              { return new MEM(e); }
    private tree.Exp plus(tree.Exp l, tree.Exp r) { return binop(PLUS, l, r); }
    private tree.Exp mult(tree.Exp l, tree.Exp r) { return binop(MULT, l, r); }
  
    private tree.Exp binop(int op, tree.Exp l, tree.Exp r)
    {
        if(l instanceof CONST && r instanceof CONST) {
            int a = ((CONST)l).value;
            int b = ((CONST)r).value;
            switch(op) {
            case PLUS:  return new CONST(a+b);
            case MINUS: return new CONST(a-b);
            case TIMES: return new CONST(a*b);
            case DIV:   return new CONST(a/b);
            case AND:   return new CONST(a&b);
            case OR:    return new CONST(a|b);
            default:    return new BINOP(op, l, r);
            }
        }
        else
            return new BINOP(op, l, r);
    }

    // allocates space, returns pointer
    private Ex malloc(tree.Exp size)
    { return extcall("malloc", new ExpList(size, null)); }
  
    // these are ok, because they simplify things
    private Ex ex(tree.Exp e)  { return new Ex(e); }
    private Ex ex(tree.Stm e)  { return new Ex((new ESEQ(e, new CONST(0)))); }
    private Nx nx(tree.Exp e)  { return new Nx(new EXP(e)); }
    private Nx nx(tree.Stm e)  { return new Nx(e); }
    private Cx cx(int op, Exp l, Exp r) { return new Cx(op,l.unEx(),r.unEx());}

    // SEMANTICS WILL CALL
    public Nx assign(tree.Exp var, tree.Exp val) 
    { return nx(new MOVE(var, val)); }

    public Ex eseq(Exp head, Exp tail) 
    { return ex(eseq(head.unNx(), tail.unEx())); }

    public Ex strcat(Exp l, Exp r)
    {
        ExpList args = null;
        args = new ExpList(r.unEx(), args);
        args = new ExpList(l.unEx(), args);
        return extcall("concat", args);
    }

    public Ex binop(int op, Exp l, Exp r)
    { return ex(new BINOP(op, l.unEx(), r.unEx())); }

    public Cx cmpop(int op, Exp l, Exp r)
    { return cx(op,l,r); }
    
    public Cx strcmp(int op, Exp l, Exp r) 
    {
        ExpList args = null;
        args = new ExpList(r.unEx(), args);
        args = new ExpList(l.unEx(), args);
        return cmpop(op, extcall("strcmp", args), ex(new CONST(0)));
    }

    public Exp ifExp(Exp c, Exp t, Exp f)
    {
        if(c==null) return nop();
        // if c is Cx, use that, else compare if c!=0
        Cx cx = (c instanceof Cx) ? (Cx)c : new Cx(NE, c.unEx(), new CONST(0));
        if(f!=null)
            return ex(cx.unEx(t.unEx(),f.unEx()));
        else
            return nx(cx.unNx(t.unNx()));
    }
    
    public Nx whileExp(Exp test, Exp body) 
    {
        if(!(test instanceof Cx)) return nop();

        Cx cxtest = (Cx)test;
        Label ltest = new Label("WHILEtest");
        Label lbody = new Label("WHILEbody");
        Label ldone = new Label("WHILEdone");
        /*
          Format:
          LABEL test
          CJUMP test, body, done

          LABEL body
          body
          JUMP test

          LABEL done
        */

        SEQ seqtest = new SEQ(new LABEL(ltest), cxtest.unNx(lbody, ldone));

        SEQ seqbody = new SEQ(new LABEL(lbody), new SEQ(body.unNx(),
                                                        new JUMP(ltest)));
        SEQ seqloop = new SEQ(new SEQ(seqtest, seqbody),
                              new LABEL(ldone));
        return nx(seqloop);
    }

    public Ex constInt(int i)
    { return ex(new CONST(i)); }

    public Ex nil()
    { return ex(new CONST(0)); }
  
    public Nx nop() 
    { return nx(new NOP()); }

    public Nx assign(Exp var, Exp val) 
    { return nx(new MOVE(var.unEx(), val.unEx())); }
  
    // ***********************************************************************
    // variable reading
    public Ex simpleVar(Level level, Access access) 
    {
        // probably not correct yet
        //        return ex(mem(plus(new TEMP(access.home.frame.FP()),
        //                 new CONST(-level.offset(access.home) +
        //                           access.offset))));
        return ex(mem(plus(level.getFP(access.home),
                           new CONST(access.offset))));
    }

    // array
    public Ex subscriptVar(Exp var, Exp index)
    { return ex(mem(plus(var.unEx(),
                         mult(index.unEx(), new CONST(wordsize))))); }

    // record field
    public Ex fieldVar(Exp var, int field)
    { return ex(mem(plus(var.unEx(), new CONST(wordsize * field)))); }

    // creating arrays/strings
    public Ex array(Exp size, Exp init) 
    {
        // first malloc space for the array
        // then go through assign the init value to each element
        //    Exp loc = malloc(mult(new CONST(WORD), size.unEx()));

        // or do as the book says

        // backwards so it comes out correct
        ExpList args = null;
        args = new ExpList(init.unEx(), args);
        args = new ExpList(size.unEx(), args);
        return extcall("initArray", args);
    }

    // create a record
    public Ex record(LinkedList inits) 
    {
        // field count
        int fc = inits.size();
        
        // allocate the space, store address in temp
        TEMP loc = new TEMP(new Temp());
        tree.Stm seq =
            assign(loc, malloc(new CONST(wordsize * fc)).unEx()).unNx();

        for(int i=0; i<fc; i++)
            seq = new SEQ(seq,
                          assign(mem(plus(loc, new CONST(wordsize * i))),
                                 ((Exp)inits.get(i)).unEx()).unNx());
        return ex(eseq(seq, loc));
    }

    // allocates a string, and returns the name for it
    public Ex string(Level level, String string) 
    {
        Label label = new Label("STR");
        frags.add(new DataFrag(level.string(string, label)));
        return ex(new NAME(label));
    }

    private Ex call(Label func, ExpList args)
    { return ex(new CALL(new NAME(func), args)); }
  
    public Ex call(Level cLevel, Level fLevel, Label func, LinkedList args,
                   boolean bStatic)
    {
        ExpList arglist = null;
        if(args.size()>0) // go backwards to make sure correct order in ExpList
            for(int i=args.size()-1; i>=0; i--)
                arglist = new ExpList(((Exp)args.get(i)).unEx(), arglist);
        if(bStatic)
            arglist = new ExpList(cLevel.getFP(fLevel), arglist);
        return call(func, arglist);
    }
  
    public Ex extcall(String func, ExpList args) 
    { return call(new Label(func, true), args); }
  
    public Ex let(Exp decs, Exp body) 
    {
        // first run through all decs, then execute body
        return ex(eseq(decs.unNx(), body.unEx()));
    }
    
    public void procEntryExit(Level level, Exp body) 
    {
        // add label for function entry into beginning of fragment
        if(body==null) body=nil();
        frags.add(new DataFrag(level.frame.label(level.label.label)));
        
        tree.Stm stmbody = new MOVE(new TEMP(frame.RV()), body.unEx());
        frags.add(new ProcFrag(stmbody, level));
    }
    
    public Fragment getFragments() 
    { return frags; }
}
